@@ -230,10 +230,9 @@ view_create(struct display *display,
 	GFile *file = NULL;
 	GError *error = NULL;
 
-	view = malloc(sizeof *view);
+	view = zalloc(sizeof *view);
 	if (view == NULL)
 		return view;
-	memset(view, 0, sizeof *view);
 
 	file = g_file_new_for_commandline_arg(filename);
 	basename = g_file_get_basename(file);
@@ -293,12 +292,13 @@ main(int argc, char *argv[])
 	int i;
 	int view_counter = 0;
 
+#if !GLIB_CHECK_VERSION(2, 35, 0)
 	g_type_init();
+#endif
 
-	argc = parse_options(view_options,
-			     ARRAY_LENGTH(view_options), argc, argv);
+	parse_options(view_options, ARRAY_LENGTH(view_options), &argc, argv);
 
-	d = display_create(argc, argv);
+	d = display_create(&argc, argv);
 	if (d == NULL) {
 		fprintf(stderr, "failed to create display: %m\n");
 		return -1;
