@@ -24,14 +24,10 @@
 #define _CAIRO_UTIL_H
 
 #include <cairo.h>
-#include <pixman.h>
 
 void
 surface_flush_device(cairo_surface_t *surface);
 
-void
-blur_surface(cairo_surface_t *surface, int margin);
-
 void
 tile_mask(cairo_t *cr, cairo_surface_t *surface,
 	  int x, int y, int width, int height, int margin, int top_margin);
@@ -90,7 +86,4 @@ enum theme_location {
 enum theme_location
 theme_get_location(struct theme *t, int x, int y, int width, int height, int flags);
 
-pixman_image_t *
-load_image(const char *filename);
-
 #endif
