@@ -11,8 +11,6 @@ extern const struct wl_interface wl_surface_interface;
 extern const struct wl_interface wl_surface_interface;
 extern const struct wl_interface wl_surface_interface;
 extern const struct wl_interface wl_output_interface;
-extern const struct wl_interface wl_surface_interface;
-extern const struct wl_interface wl_output_interface;
 
 static const struct wl_interface *types[] = {
 	NULL,
@@ -28,8 +26,6 @@ static const struct wl_interface *types[] = {
 	NULL,
 	&wl_surface_interface,
 	&wl_output_interface,
-	&wl_surface_interface,
-	&wl_output_interface,
 };
 
 static const struct wl_message desktop_shell_requests[] = {
@@ -38,17 +34,18 @@ static const struct wl_message desktop_shell_requests[] = {
 	{ "set_lock_surface", "o", types + 5 },
 	{ "unlock", "", types + 0 },
 	{ "set_grab_surface", "o", types + 6 },
+	{ "desktop_ready", "2", types + 0 },
 };
 
 static const struct wl_message desktop_shell_events[] = {
-	{ "configure", "uoii", types + 7 },
-	{ "prepare_lock_surface", "", types + 0 },
-	{ "grab_cursor", "u", types + 0 },
+	{ "configure", "2uoii", types + 7 },
+	{ "prepare_lock_surface", "2", types + 0 },
+	{ "grab_cursor", "2u", types + 0 },
 };
 
 WL_EXPORT const struct wl_interface desktop_shell_interface = {
-	"desktop_shell", 1,
-	5, desktop_shell_requests,
+	"desktop_shell", 2,
+	6, desktop_shell_requests,
 	3, desktop_shell_events,
 };
 
@@ -62,13 +59,3 @@ WL_EXPORT const struct wl_interface screensaver_interface = {
 	0, NULL,
 };
 
-static const struct wl_message input_panel_requests[] = {
-	{ "set_surface", "oo", types + 13 },
-};
-
-WL_EXPORT const struct wl_interface input_panel_interface = {
-	"input_panel", 1,
-	1, input_panel_requests,
-	0, NULL,
-};
-
