@@ -21,7 +21,7 @@ static const struct wl_message xserver_events[] = {
 
 WL_EXPORT const struct wl_interface xserver_interface = {
 	"xserver", 1,
-	ARRAY_LENGTH(xserver_requests), xserver_requests,
-	ARRAY_LENGTH(xserver_events), xserver_events,
+	1, xserver_requests,
+	2, xserver_events,
 };
 
